"""
Boston Micromachines, deformable mirrors.

The qmi.instruments.boston_micromachines package supports Multi-DM models with 140 actuators.
"""

from qmi.instruments.boston_micromachines.multidm import BostonMicromachines_MultiDM
from qmi.instruments.boston_micromachines.multidm import write_mirror_shape, read_mirror_shape
