"""
Time Base, AOM driver.

The qmi.instruments.timebase package provides support for:
- DIM3000 AOM driver.
"""
# Alternative, QMI naming convention approved names
from qmi.instruments.timebase.dim3000 import TimeBase_DIM3000 as TimeBase_Dim3000
