"""
Quantum Opus, low-noise amplifier module.

The qmi.instruments.quantum_opus package provides support for:
- QO-AMP-SIM module.
"""
# Alternative, QMI naming convention approved names
from qmi.instruments.quantum_opus.amp_sim_module import AmpSimModule as QuantumOpus_QoAmpSim
