"""
Rohde & Schwarz, analog and vector signal generators.

The qmi.instruments.rohde_schwarz package provides support for:
- SGS100A model
- SMBV100A model.
"""
# Alternative, QMI naming convention approved names
from qmi.instruments.rohde_schwarz.sgs100a import RohdeSchwarz_SGS100A as RohdeSchwarz_Sgs100a
from qmi.instruments.rohde_schwarz.smbv100a import RohdeSchwarz_SMBV100A as RohdeSchwarz_Smbv100a
