"""
Measurement Computing Corporation (MCC), DAQ devices.

The qmi.instruments.mcc package provides limited support for:
- USB-1808 series devices.
"""
# Alternative, QMI naming convention approved names
from qmi.instruments.mcc.usb1808x import MCC_USB1808X as MCC_Usb1808x
