"""
Hübner Photonics, diode laser.

The qmi.instruments.cobolt package provides support for:
- Continuous Wave 06-01 series diode laser.
"""
# Alternative, QMI naming convention approved name
from qmi.instruments.cobolt.laser_06_01 import Cobolt_Laser_06_01 as Hubner_Cobolt0601
