"""
Teraxion, narrowband tunable optical filters.

The qmi.instruments.teraxion package provides support for:
  - Teraxion TFN model (tested).
"""
from qmi.instruments.teraxion.tfn import Teraxion_TFN
