"""
Raspberry Pi, computer modules.

The qmi.instruments.raspberry package provides support for:
- GPIO header of various computer models.
"""
# Alternative, QMI naming convention approved names
from qmi.instruments.raspberry.gpio import RaspberryPiGPIO as RaspberryPi_Gpio
