"""
Imagine Eyes, deformable mirror.

The qmi.instruments.imagine_eyes package provides support for:
- Mirao 52-e electromagnetic deformable mirror.
"""

from qmi.instruments.imagine_eyes.mirao52e import ImagineEyes_Mirao52e, MirrorStatus, read_mirao_file
