"""
Anapico, signal generators.

The qmi.instruments.agiltron package provides support for:
- APSIN Series signal generators.
"""
# Alternative, QMI naming convention approved name
from qmi.instruments.anapico.apsin import Anapico_APSIN as Anapico_Apsin
