"""
Agiltron, optical switch driver device.

The qmi.instruments.agiltron package provides support for:
- FF 1x4 optical switch driver
"""
# Using QMI naming convention approved name
from qmi.instruments.agiltron.ff1x4 import Agiltron_FF1x4 as Agiltron_Ff1x4
