"""
High Finesse, Wavelength meter.

The qmi.instruments.high_finesse package provides support for:
  - HighFinesse WLM Controller model WS-6 (tested).
  - HighFinesse WLM Controller model WS-5, WS-7, WS-8 (untested).
"""
from qmi.instruments.high_finesse.wlm import HighFinesse_Wlm
