"""
InstruTech Inc., vacuum gauge controller.

The qmi.instruments.instru_tech package provides support for:
- AGC302 active vacuum gauge controller.
"""
# Alternative, QMI naming convention approved name
from qmi.instruments.instru_tech.instrutech_agc302 import InstruTech_AGC302 as InstruTech_Agc302
