"""
Tenma (Newark Electronics) power supply units.

The qmi.instruments.tenma package provides support for:
- Series 72 power supply unit groups
  - 72-2535, 72-2540, 72-2545, 72-2550, 72-2925, 72-2930, 72-2935, 72-2940 & 72-10480
  - 72-13350 & 72-13360
"""
from qmi.instruments.tenma.psu_72 import *
